
#include "sys.h"
#include <stdio.h>
#include <stdlib.h>

#ifdef WINDOWS
	#include <shlobj.h>
	#define SEP L'\\'
	#define ALT_SEP L'/'
#else
	#include <errno.h>
	#include <sys/stat.h>
	#include <unistd.h>
	#include <dirent.h>
	#define SEP '/'
	#define ALT_SEP '\\'
#endif

struct File
{
	#ifdef WINDOWS
		HANDLE handle;
	#else
		FILE* handle;
	#endif
};

struct FileStat
{
	int64_t	size;
	uint64_t	lastModification;
	char	isReadOnly;
	char	isDirectory;
};

// Opaque pointer
struct Directory
{
	#ifdef WINDOWS
		WIN32_FIND_DATA	data;
		HANDLE	handle;
	#else
		DIR* dir;
		struct dirent* dirent;
	#endif
};

// FileStat
//[of]:fileStatSet
// Get file information
//
// RETURN VALUE
//	OK if success, error otherwise.
//
ENTRY SysErrorCode csys_fileStatSet(struct FileStat* fs, const String path)
{
	#ifdef WINDOWS
		SysErrorCode res;
		HANDLE h;
		wchar* filename;
		BOOL success;
		LARGE_INTEGER  size;
		FILETIME ft;
		DWORD attributes;
		
		res = csys_newWideFromString(&filename, path);
		if( res != OK )
			return res;
	
		attributes = GetFileAttributes(filename);
		if( attributes == INVALID_FILE_ATTRIBUTES )
		{
			res = ERR_SYSTEM;
		}
		else
		{
			h = CreateFile (	filename,
				GENERIC_READ,
				FILE_SHARE_READ|FILE_SHARE_WRITE,
				0,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS,
				0);
			if( h == INVALID_HANDLE_VALUE )
			{
				res = ERR_SYSTEM;
			}
			else
			{
				success = GetFileSizeEx(h, &size);
				if( success == 0 )
				{
					res = ERR_SYSTEM;
				}
				else
				{
					success = GetFileTime (h, NULL, NULL, &ft);
					if ( success == 0 )
					{
						res = ERR_SYSTEM;
					}
					else
					{
						fs->isReadOnly = (attributes & FILE_ATTRIBUTE_READONLY) != 0;
						fs->isDirectory = (attributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
						fs->lastModification = csys_windowsTimeToTime((((uint64_t)ft.dwHighDateTime) << 32) + (uint64_t)ft.dwLowDateTime);
						fs->size = size.QuadPart;
						res = OK;
					}
				}
				// We assume that the last error is not erased in case of success
				success = CloseHandle (h);
				if( success == 0 )
					res = ERR_SYSTEM;
			}
		}
			
		csys_deleteWide(filename);
		return res;
	
	#else
	
		SysErrorCode res;
		struct stat st;
		char* filename;
		int err;
		char ro;
		uint64_t mtime;
		
		res = csys_newLocaleFromString(&filename, path);
		if( res != OK )
			return res;
		
		err = stat(filename, &st);
		if( err != 0 )
		{
			res = ERR_SYSTEM;
		}
		else
		{
			err = access(filename, W_OK);
			
			// Whatever happened, an error is considered as a read only file.
			// Technically, only EACCESS and EROFS should be considered,
			// other error should be reported, even ENOENT can happen if the
			// file is deleted between stat() and access().
			ro = err != 0;
			
			fs->isReadOnly = ro;
			fs->size = (int64_t)st.st_size;
			fs->isDirectory = (char)S_ISDIR(st.st_mode);
			fs->lastModification = csys_unixTimeToTime((uint64_t)st.st_mtime);
	
			res = OK;
		}
	
		csys_deleteLocale(filename);
		return res;
	
	#endif
}
//[cf]

// File
//[of]:fileExists
ENTRY SysErrorCode csys_fileExists(const String path, int* exists)
{
	#ifdef WINDOWS
	
		wchar* filename;
		DWORD attributes;
		
		SysErrorCode res = csys_newWideFromString(&filename, path);
		if( res != OK )
			return res;
		attributes = GetFileAttributes (filename);
		csys_deleteWide (filename);
		if( attributes == INVALID_FILE_ATTRIBUTES )
			*exists = 0;
		else
			*exists = (attributes & FILE_ATTRIBUTE_DIRECTORY) == 0;
		return OK;
	
	#else
	
		struct stat st;
		SysErrorCode res;
		int err;
		char* filename;
		
		res = csys_newLocaleFromString(&filename, path);
		if( res != OK )
			return res;
		err = stat(filename, &st);
		csys_deleteLocale(filename);
		
		if( err == 0 )
		{
			*exists = !S_ISDIR(st.st_mode);
			res = OK;
		}
		else if( errno == ENOENT )
		{
			*exists = 0;
			res = OK;
		}
		else
			res = ERR_SYSTEM;
		return res;
	
	
	#endif
}
//[cf]
//[of]:filePathIsAbsolute
ENTRY int csys_filePathIsAbsolute(String filename)
{
	#ifdef WINDOWS
		// Already a fully qualified name ?
		return	wcslen(filename) > 1 && (filename[1] == L':' || (filename[0] == L'\\' && filename[1] == L'\\'));
	#else
		return	filename[0] == '/';
	#endif
}
//[cf]
//[of]:fileNormalize
//[c]Normalizes a full name.
//[c]* reduces '\.'
//[c]* reduces '\..'
//[c]* reduces '\\' (except initial)
//[c]
//[c]PARAMETERS
//[c]	fullName	--	A fully qualified name, i.e. it must starts with '\\' or 'x:\'.
//[c]
//[c]RETURN VALUES
//[c]	The new end of the string.
//[c]
ENTRY String csys_fileNormalize(String fullName)
{
	String s = fullName;
	String d;
	String start;

	#ifdef WINDOWS
		if( s[0] == SEP && s[1] == SEP )
			s++;
		else
			// skip unit
			s+= 2;
	#endif
	
	d = s;
	start = d;
		
	for(;;)
	{
		CodeUnit c = *s++;
		if( c == 0 )
			break;
		if( c == SEP )
		{
			// Skip adjacent separators
			while( *s == SEP )
				s++;
			if( s[0] == '.' && (s[1] == 0 || s[1] == SEP) )
				s++;
			else if( s[0] == '.' && s[1] == '.' && (s[2] == 0 || s[2] == SEP) )
			{
				String p = d - 1;
				s += 2;
				while( p >= start )
				{
					if( *p == SEP )
					{
						d = p;
						break;
					}
					p--;
				}
			}
			else
			{
				*d++ = c;
			}
		}
		else
		{
			*d++ = c;
		}
	}

	// Empty: restore the initial '\' (both UNC and x:\)
	if( d == start )
		*d++ = SEP;

	return d;
}
//[cf]
//[of]:fileIsCaseSensitive
ENTRY int csys_fileIsCaseSensitive()
{
#ifdef WINDOWS
	return 0;
#else
	return 1;
#endif
}
//[cf]
//[of]:fileDestroy
//[c]Deletes the file.
//[c]
ENTRY SysErrorCode csys_fileDestroy (String input)
{
	#ifdef WINDOWS
		int res;
		BOOL success;
		wchar* filename;
		
		res = csys_newWideFromString(&filename, input);
		if( res != OK )
			return res;
		
		success = DeleteFile(filename);
		csys_deleteWide(filename);
		return ( success != 0 ) ? OK : ERR_SYSTEM;
	#else
		int res;
		int err;
		char* filename;
		
		res = csys_newLocaleFromString(&filename, input);
		if( res != OK )
			return res;
		
		err = unlink(filename);
		csys_deleteLocale(filename);
		return ( err == 0 ) ? OK : ERR_SYSTEM;
	#endif
}
//[cf]
//[of]:fileOpen
ENTRY SysErrorCode csys_fileOpen(struct File* file, String filename, int mode)
{
	#ifdef WINDOWS
		HANDLE h = CreateFile (	filename,
			(mode != OPEN_WRITE) ? GENERIC_READ : GENERIC_WRITE,
			(mode != OPEN_WRITE) ? FILE_SHARE_READ|FILE_SHARE_WRITE : 0,
			NULL,
			(mode != OPEN_WRITE) ? OPEN_EXISTING : CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL);
		if( h == INVALID_HANDLE_VALUE )
			return ERR_SYSTEM;
		file->handle = h;
		return OK;
	#else
		int res;
		char* s;
		const char* modestr;
		FILE* h;
		
		res = csys_newLocaleFromString(&s, filename);
		if( res != OK )
			return res;
	
		if( mode == OPEN_WRITE )
			modestr = "w";
		else
			modestr = "r";
			
		h = fopen (s, modestr);
		csys_deleteLocale(s);
		if( h == 0 )
			return ERR_SYSTEM;
		
		file->handle = h;
		return OK;
	#endif
}
//[cf]
//[of]:fileClose
ENTRY SysErrorCode csys_fileClose(struct File* file)
{
	#ifdef WINDOWS
		DWORD res = CloseHandle(file->handle);
		if( res == 0 )
			return ERR_SYSTEM;
		file->handle = INVALID_HANDLE_VALUE;
		return OK;
	#else
		int res = fclose (file->handle);
		if (res != 0)
			return ERR_SYSTEM;
		file->handle = 0;
		return OK;
	#endif
}
//[cf]
//[of]:fileRead
ENTRY SysErrorCode csys_fileRead(struct File* file, void* base, size_t size, size_t* read)
{
	#ifdef WINDOWS
		DWORD n;
		// The size is limited to 4GB		
		DWORD res = ReadFile (file->handle, base, (DWORD)size, &n, NULL);
		if( res == 0 )
			return ERR_SYSTEM;
		*read = (size_t)n;
		return OK;
	#else
		int n = fread(base, 1, size, file->handle);
		if( n != size && ferror (file->handle) != 0 )
			return ERR_SYSTEM;
		*read = n;
		return OK;
	#endif
}
//[cf]
//[of]:fileWrite
ENTRY SysErrorCode csys_fileWrite(struct File* file, void* base, size_t size, size_t* written)
{
	#ifdef WINDOWS
		DWORD n;
		// The size is limited to 4GB		
		DWORD res = WriteFile(file->handle, base, (DWORD)size, &n, NULL);
		if( res == 0 )
			return ERR_SYSTEM;
		*written = n;
		return OK;
	#else
		int n = fwrite(base, 1, size, file->handle);
		if( n != size && ferror (file->handle) != 0 )
			return ERR_SYSTEM;
		*written = n;
		return OK;
	#endif
}
//[cf]
//[of]:fileSeek
ENTRY SysErrorCode csys_fileSeek(struct File* file, size_t size, SeekMode mode)
{
	#ifdef WINDOWS
		DWORD res;
		DWORD dwMethod;
		if( mode == SM_SET )
			dwMethod = FILE_BEGIN;
		else if( mode == SM_CURRENT )
			dwMethod = FILE_CURRENT;
		else
			dwMethod = FILE_END;
		
		res = SetFilePointer(file->handle, (LONG)size, NULL, dwMethod);
		if( res == INVALID_SET_FILE_POINTER && GetLastError() != NO_ERROR )
			return ERR_SYSTEM;
		return OK;
	#else
		int origin;
		if( mode == SM_SET )
			origin = SEEK_SET;
		else if( mode == SM_CURRENT )
			origin = SEEK_CUR;
		else
			origin = SEEK_END;
		
		int res = fseek(file->handle, (long int)size, origin);
		return ( res != 0 ) ? ERR_SYSTEM : OK;
	#endif
}
//[cf]
//[of]:getPathSeparator
// Copper can't import a variable, let's use a function
ENTRY int csys_getPathSeparator()
{
	return SEP;
}
//[cf]
//[of]:getAltPathSeparator
// Copper can't import a variable, let's use a function
ENTRY int csys_getAltPathSeparator()
{
	return ALT_SEP;
}
//[cf]
//[of]:getNewLine
// Copper can't import a variable, let's use a function
ENTRY const String csys_getNewLine()
{
	#ifdef WINDOWS
		return L"\r\n";
	#else
		return "\n";
	#endif
}
//[cf]

// Directory
//[of]:dirCreate
ENTRY int csys_dirCreate(String input)
{
	#ifdef WINDOWS
		int res;
		int success;
		wchar* path;
		
		res = csys_newWideFromString(&path, input);
		if( res != OK )
			return res;
		success = CreateDirectory(path, 0);
		csys_deleteWide(path);
		return ( success != 0 ) ? OK : ERR_SYSTEM;
	#else
		int res;
		int err;
		char* path;
		unsigned int mode;
		
		res = csys_newLocaleFromString(&path, input);
		if( res != OK )
			return res;
		mode = (7 << 6) | (5 << 3) | 5;
		err = mkdir(path, mode);
		csys_deleteLocale(path);
		return (err == 0) ? OK : ERR_SYSTEM;
	#endif
}
//[cf]

//[of]:dirNext
ENTRY int csys_dirNext(struct Directory* directory, int* hasMore)
{
	#ifdef WINDOWS

		BOOL success = FindNextFile(directory->handle, &directory->data);
		if( success == 0 )
		{
			if( GetLastError() == ERROR_FILE_NOT_FOUND )
				*hasMore = 0;
			else
				return ERR_SYSTEM;
		}
		else
			*hasMore = 1;
		return OK;

	#else

		directory->dirent = readdir(directory->dir);
		if( directory->dirent == NULL )
			*hasMore = 0;
		else
			*hasMore = 1;
		return OK;

	#endif
}
//[cf]
//[of]:dirClose
ENTRY int csys_dirClose(struct Directory* directory)
{
	#ifdef WINDOWS
		if( directory->handle != INVALID_HANDLE_VALUE )
		{
			BOOL success = FindClose(directory->handle);
			directory->handle = INVALID_HANDLE_VALUE;
			return (success != 0) ? OK : ERR_SYSTEM;
		}
	#else
		if( directory->dir != NULL )
		{
			int err = closedir(directory->dir);
			directory->dir = NULL;
			return (err == 0) ? OK : ERR_SYSTEM;
		}
	#endif
	
	return OK;
}
//[cf]
//[of]:dirNewName
//[c]Gets the name for the current entry.
//[c]
//[c]If the function succeeds, the name must be deleted with csys_deleteString.
//[c]
ENTRY SysErrorCode csys_dirNewName(struct Directory* directory, String* out)
{
	#ifdef WINDOWS
		return csys_newStringFromWide(out, directory->data.cFileName);
	#else
		return csys_newStringFromLocale(out, directory->dirent->d_name);
	#endif
}
//[cf]
//[of]:dirIsFile
ENTRY int csys_dirIsFile(struct Directory* directory)
{
	#ifdef WINDOWS
		const wchar* n = directory->data.cFileName;
		if( n[0] == '.' )
		{
			if( n[1] == 0 )
				// .
				return 0;
			else if( n[1] == '.' && n[2] == 0 )
				// ..
				return 0;
		}
		return (directory->data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0;
	#else
		
		const char* n = directory->dirent->d_name;
		if( n[0] == '.' )
		{
			if( n[1] == 0 )
				// .
				return 0;
			else if( n[1] == '.' && n[2] == 0 )
				// ..
				return 0;
		}
		return directory->dirent->d_type != DT_DIR;
	#endif
}
//[cf]
//[of]:dirIsDirectory
ENTRY int csys_dirIsDirectory(struct Directory* directory)
{
	#ifdef WINDOWS
		const wchar* n = directory->data.cFileName;
		if( n[0] == '.' )
		{
			if( n[1] == 0 )
				// .
				return 0;
			else if( n[1] == '.' && n[2] == 0 )
				// ..
				return 0;
		}
		return (directory->data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
	#else
		
		const char* n = directory->dirent->d_name;
		if( n[0] == '.' )
		{
			if( n[1] == 0 )
				// .
				return 0;
			else if( n[1] == '.' && n[2] == 0 )
				// ..
				return 0;
		}
		return directory->dirent->d_type == DT_DIR;
	#endif
}
//[cf]
//[of]:dirDelete
ENTRY void csys_dirDelete(struct Directory* directory)
{
	csys_dirClose(directory);
	csys_free(directory, sizeof(struct Directory));
}
//[cf]
//[of]:dirNew
ENTRY SysErrorCode csys_dirNew(struct Directory** out, String path, int* hasMore)
{
	int res;
	struct Directory* directory;
	
	#ifdef WINDOWS
	
		size_t pathlen;
		wchar* dirname;
		wchar* mask;
		HANDLE handle;
		
		// Initialize
		directory = (struct Directory*)csys_allocate(sizeof(struct Directory));
		directory->handle = INVALID_HANDLE_VALUE;

		res = csys_newWideFromString(&dirname, path);
		if( res == OK )
		{
			pathlen = wcslen(dirname);
			mask = (wchar*)csys_allocate((pathlen + 4 + 1) * sizeof(wchar));
			memcpy(mask, path, pathlen * sizeof(wchar));
			memcpy(mask + pathlen, L"\\*.*", (4 + 1) * sizeof(wchar));
		
			directory->handle = FindFirstFile(mask, &directory->data);
			if( directory->handle == INVALID_HANDLE_VALUE )
			{
				if( GetLastError() == ERROR_FILE_NOT_FOUND )
					*hasMore = 0;
				else
					res = ERR_SYSTEM;
			}
			else
			{
				*hasMore = 1;
			}
			csys_free(mask, (pathlen + 4 + 1) * sizeof(wchar));
			csys_deleteWide(dirname);
		}
		
	#else
	
		char* dirname;
		DIR* dir;
	
		// Initialize
		directory = (struct Directory*)csys_allocate(sizeof(struct Directory));
		directory->dir = NULL;
	
		csys_newLocaleFromString(&dirname, path);
		dir = opendir(dirname);
		if( dir == NULL )
		{
			res = ERR_SYSTEM;
		}
		else
		{
			directory->dir = dir;
			res = csys_dirNext(directory, hasMore);
		}
		csys_deleteLocale(dirname);
	
	#endif

	if( res != OK )
		csys_dirDelete(directory);
	else
		*out = directory;
	return res;
}
//[cf]

//[of]:dirAddCurrentDirectoryTo
//[c]Returns the current directory. 
//[c]
//[c]If the function succeeds, the string must be deleted with csys_deleteString.
//[c]
ENTRY SysErrorCode csys_dirAddCurrentDirectoryTo(StringBuffer* out)
{
	#ifdef WINDOWS
		wchar buf[1024];
		if( GetCurrentDirectory (1024, buf) == 0 )
			return ERR_SYSTEM;
		
		return StringBuffer_addWide(out, buf);
	#else
		char buf[1024];
		if( getcwd (buf, 1024) == 0 )
			return ERR_SYSTEM;
		
		return StringBuffer_addLocale(out, buf);
	#endif
}
//[cf]
//[of]:dirAddApplicationDirectoryTo
//[c]Returns the directory name to store data for an application named 'name'
//[c]for all users. This directory is considered read-only.
//[c]
//[c]PARAMETERS
//[c]	out	--	[out] The pointer to the resulting string.
//[c]	name	--	The name of the application. Used only on Unix system.
//[c]	datadir	--	'/usr/share' or '/urs/local/share'. Used only on Unix system.
//[c]
//[c]RETURN VALUES
//[c]	A newly allocated string containing the application directory.
//[c]	
//[c]REMARKS
//[c]	On Windows it corresponds to the directory of the current executable, on 
//[c]	Linux, it corresponds to '/usr/share/<name>' or '/usr/local/share/<name>'.
//[c]	
ENTRY SysErrorCode csys_dirAddApplicationDirectoryTo (StringBuffer* out, String name, const String datadir)
{
	#ifdef WINDOWS
		wchar* s;
		
		// Get the full name of the executable program
		wchar buf[MAX_PATH];
		GetModuleFileName (NULL, buf, MAX_PATH);
		
		// Get the path of the executable program
		// 1. find the end of the string
		s = buf;
		while( *s != 0 ) 
			s++;
		// 2. find the last separator
		while( s != buf )
		{
			s--;
			if( *s == '\\' )
			{
				// truncate
				*s = 0;
				break;
			}
		}
		
		StringBuffer_addWide(out, buf);
		return OK;
	#else
		// Unix: create a string with the format
		//	datadir / name
		//
		StringBuffer_addString(out, datadir);
		StringBuffer_addString(out, "/");
		StringBuffer_addString(out, name);
		return OK;
	#endif
}
//[cf]
//[of]:dirAddUserDirectoryTo
//[c]Returns the directory name to store data for an application named 'name'
//[c]for the current user.
//[c]
//[c]PARAMETERS
//[c]	out	--	The buffer where the result is appended.
//[c]	name	--	The name of the application.
//[c]
//[c]REMARKS
//[c]	The directory may not exist.
//[c]
ENTRY SysErrorCode csys_dirAddUserDirectoryTo (StringBuffer* out, String name)
{
	SysErrorCode res = OK;

	#ifdef WINDOWS
		// Get the AppData path
		wchar buf[MAX_PATH];
		BOOL success = SHGetSpecialFolderPath(NULL, buf, CSIDL_APPDATA, FALSE);
		if( success == 0 )
			res = ERR_SYSTEM;
		if( res == OK )
			res = StringBuffer_addWide(out, buf);
		if( res == OK )
			StringBuffer_addString(out, L"\\");
	#else
		
		char* cfg = getenv("XDG_CONFIG_HOME");
		if( cfg != NULL && *cfg != 0 )
		{
			res = StringBuffer_addLocale(out, cfg);
		}
		else
		{
			char* home = getenv ("HOME");
			if( home != NULL )
				res = StringBuffer_addLocale(out, home);
			if( res == OK )
			 	StringBuffer_addString(out, "/.config");
		}
		if( res == OK )
			StringBuffer_addString(out, "/");

	#endif

	if( res == OK )
		StringBuffer_addString(out, name);
	return res;
}
//[cf]
