
#include "sys.h"
#ifdef WINDOWS
//[of]:	thread
typedef struct WrapStartRoutine
{
	void *(*func)(void *);
	void* arg;
 } WrapStartRoutine;

DWORD WINAPI startRoutine(LPVOID p)
{
	((WrapStartRoutine*)p)->func(((WrapStartRoutine*)p)->arg);
	return 0;
}

SysErrorCode csys_threadCreate(HANDLE* out, void *(*func)(void *), void* arg)
{
	WrapStartRoutine wrap;
	wrap.func = func;
	wrap.arg = arg;
	
	HANDLE thread = CreateThread(
		NULL,
		0,
		startRoutine,
		&wrap,
		0,
		NULL);
	
	if( thread == NULL )
		return ERR_SYSTEM;
	
	*out = thread;
	return OK;
}

int csys_threadJoin(HANDLE thread)
{
	WaitForSingleObject(thread, INFINITE);
	return OK;
}

int csys_threadKill(HANDLE thread)
{
	BOOL res = TerminateThread(thread, 0);
	return res ? 0 : 1;
}
//[cf]
//[of]:	mutex
SysErrorCode csys_mutexCreate(LPCRITICAL_SECTION* out)
{
	LPCRITICAL_SECTION mutex = (LPCRITICAL_SECTION)csys_allocate(sizeof(CRITICAL_SECTION));
	InitializeCriticalSection(mutex);
	*out = mutex;
	return OK;
}

int csys_mutexDestroy(LPCRITICAL_SECTION mutex)
{
	DeleteCriticalSection(mutex);
	csys_free(mutex, sizeof(CRITICAL_SECTION));
	return 0;
}

int csys_mutexLock(LPCRITICAL_SECTION mutex)
{
	EnterCriticalSection(mutex);
	return 0;
}

int csys_mutexUnlock(LPCRITICAL_SECTION mutex)
{
	LeaveCriticalSection(mutex);
	return 0;
}
//[cf]
//[of]:	cond
SysErrorCode csys_condCreate(PCONDITION_VARIABLE* out)
{
	PCONDITION_VARIABLE cond = (PCONDITION_VARIABLE)csys_allocate(sizeof(CONDITION_VARIABLE));
	InitializeConditionVariable(cond);
	*out = cond;
	return OK;
}

int csys_condDestroy(PCONDITION_VARIABLE cond)
{
	csys_free(cond, sizeof(CONDITION_VARIABLE));
	return 0;
}

int csys_condWait(PCONDITION_VARIABLE cond, LPCRITICAL_SECTION mutex)
{
	BOOL res = SleepConditionVariableCS(cond, mutex, INFINITE);
	return res ? 0 : 1;
}

int csys_condBroadcast(PCONDITION_VARIABLE cond)
{
	WakeAllConditionVariable(cond);
	return 0;
}

int csys_condSignal(PCONDITION_VARIABLE cond)
{
	WakeConditionVariable(cond);
	return 0;
}
//[cf]

#else
	#include <stddef.h>
	#include <stdlib.h>
	#include <pthread.h>
	#include <signal.h>

//[of]:	thread
void initializeLocale();

typedef struct Thread
{
	pthread_t handle;
} Thread;

SysErrorCode csys_threadCreate(Thread** out, void *(*func)(void *), void* arg)
{
	// The text module is not thread-safe: the initialization must be
	// done before the creation of a new thread to ensure that two threads
	// won't attempt to do the initialization at the same time.
	initializeLocale();
	
	pthread_t handle;
	int res = pthread_create (&handle, NULL, func, arg);
	if( res == 0 )
	{
		Thread* p = (Thread*)csys_allocate(sizeof(Thread));
		p->handle = handle;
		*out = p;
		return OK;
	}
	return ERR_SYSTEM;
}

int csys_threadJoin(Thread* thread)
{
	void* retval;
	int res = pthread_join(thread->handle, &retval);
	if( res == 0 )
		csys_free(thread, sizeof(Thread));
	return res;
}

int csys_threadKill(Thread* thread)
{
	return pthread_kill(thread->handle, SIGKILL);
}
//[cf]
//[of]:	mutex
SysErrorCode csys_mutexCreate(pthread_mutex_t** out)
{
  pthread_mutex_t* mutex = (pthread_mutex_t*)csys_allocate(sizeof(pthread_mutex_t));
  int res = pthread_mutex_init(mutex, NULL);
  if( res != 0 )
  {
    csys_free(mutex, sizeof(pthread_mutex_t));
    return ERR_SYSTEM;
  }
  *out = mutex;
  return OK;
}

int csys_mutexDestroy(pthread_mutex_t* mutex)
{
  int res = pthread_mutex_destroy(mutex);
  if( res == 0 )
    csys_free(mutex, sizeof(pthread_mutex_t));
  return res;
}

int csys_mutexLock(pthread_mutex_t* mutex)
{
  return pthread_mutex_lock(mutex);
}

int csys_mutexUnlock(pthread_mutex_t* mutex)
{
  return pthread_mutex_unlock(mutex);
}
//[cf]
//[of]:	cond
SysErrorCode csys_condCreate(pthread_cond_t** out)
{
  pthread_cond_t* cond = (pthread_cond_t*)csys_allocate(sizeof(pthread_cond_t));
  int res = pthread_cond_init(cond, NULL);
  if( res != 0 )
  {
    csys_free(cond, sizeof(pthread_cond_t));
    return ERR_SYSTEM;
  }
  *out = cond;
  return OK;
}

int csys_condDestroy(pthread_cond_t* cond)
{
  int res = pthread_cond_destroy(cond);
  if( res == 0 )
    csys_free(cond, sizeof(pthread_cond_t));
  return res;
}

int csys_condWait(pthread_cond_t* cond, pthread_mutex_t* mutex)
{
  return pthread_cond_wait(cond, mutex);
}

int csys_condBroadcast(pthread_cond_t* cond)
{
  return pthread_cond_broadcast(cond);
}

int csys_condSignal(pthread_cond_t* cond)
{
  return pthread_cond_signal(cond);
}
//[cf]
#endif
